@@ -7,6 +7,8 @@
 #ifndef TREELITE_PREDICTOR_H_
 #define TREELITE_PREDICTOR_H_
 
+#include <treelite/error.h>
+#include <treelite/omp_exception.h>
 #include <treelite/logging.h>
 #include <treelite/typeinfo.h>
 #include <treelite/c_api_runtime.h>
@@ -16,48 +18,14 @@
 #include <mutex>
 #include <cstdint>
 
+#ifdef _WIN32
+#define NOMINMAX
+#include <windows.h>
+#endif  // _WIN32
+
 namespace treelite {
 namespace predictor {
 
-/*!
- * \brief OMP Exception class catches, saves and rethrows exception from OMP blocks
- */
-class OMPException {
- private:
-  // exception_ptr member to store the exception
-  std::exception_ptr omp_exception_;
-  // mutex to be acquired during catch to set the exception_ptr
-  std::mutex mutex_;
-
- public:
-  /*!
-   * \brief Parallel OMP blocks should be placed within Run to save exception
-   */
-  template <typename Function, typename... Parameters>
-  void Run(Function f, Parameters... params) {
-    try {
-      f(params...);
-    } catch (treelite::Error &ex) {
-      std::lock_guard<std::mutex> lock(mutex_);
-      if (!omp_exception_) {
-        omp_exception_ = std::current_exception();
-      }
-    } catch (std::exception &ex) {
-      std::lock_guard<std::mutex> lock(mutex_);
-      if (!omp_exception_) {
-        omp_exception_ = std::current_exception();
-      }
-    }
-  }
-
-  /*!
-   * \brief should be called from the main thread to rethrow the exception
-   */
-  void Rethrow() {
-    if (this->omp_exception_) std::rethrow_exception(this->omp_exception_);
-  }
-};
-
 /*! \brief data layout. The value -1 signifies the missing value.
     When the "missing" field is set to -1, the "fvalue" field is set to
     NaN (Not a Number), so there is no danger for mistaking between
@@ -71,8 +39,13 @@ union Entry {
 
 class SharedLibrary {
  public:
+#ifdef _WIN32
+  using LibraryHandle = HMODULE;
+  using FunctionHandle = FARPROC;
+#else  // _WIN32
   using LibraryHandle = void*;
   using FunctionHandle = void*;
+#endif  // _WIN32
   SharedLibrary();
   ~SharedLibrary();
   void Load(const char* libpath);
@@ -199,6 +172,13 @@ class Predictor {
   inline float QuerySigmoidAlpha() const {
     return sigmoid_alpha_;
   }
+  /*!
+   * \brief Get c value in exponential standard ratio used to train the loaded model
+   * \return c value in exponential standard ratio transformation
+   */
+  inline float QueryRatioC() const {
+    return ratio_c_;
+  }
   /*!
    * \brief Get global bias which adjusting predicted margin scores
    * \return global bias
@@ -240,6 +220,7 @@ class Predictor {
   size_t num_feature_;
   std::string pred_transform_;
   float sigmoid_alpha_;
+  float ratio_c_;
   float global_bias_;
   int num_worker_thread_;
   TypeInfo threshold_type_;
