@@ -7,6 +7,9 @@
 #ifndef TREELITE_TREE_IMPL_H_
 #define TREELITE_TREE_IMPL_H_
 
+#include <treelite/error.h>
+#include <treelite/version.h>
+#include <treelite/logging.h>
 #include <algorithm>
 #include <limits>
 #include <memory>
@@ -87,7 +90,7 @@ ContiguousArray<T>::Clone() const {
   ContiguousArray clone;
   clone.buffer_ = static_cast<T*>(std::malloc(sizeof(T) * capacity_));
   if (!clone.buffer_) {
-    throw std::runtime_error("Could not allocate memory for the clone");
+    throw Error("Could not allocate memory for the clone");
   }
   std::memcpy(clone.buffer_, buffer_, sizeof(T) * size_);
   clone.size_ = size_;
@@ -160,11 +163,11 @@ template <typename T>
 inline void
 ContiguousArray<T>::Reserve(std::size_t newsize) {
   if (!owned_buffer_) {
-    throw std::runtime_error("Cannot resize when using a foreign buffer; clone first");
+    throw Error("Cannot resize when using a foreign buffer; clone first");
   }
   T* newbuf = static_cast<T*>(std::realloc(static_cast<void*>(buffer_), sizeof(T) * newsize));
   if (!newbuf) {
-    throw std::runtime_error("Could not expand buffer");
+    throw Error("Could not expand buffer");
   }
   buffer_ = newbuf;
   capacity_ = newsize;
@@ -174,7 +177,7 @@ template <typename T>
 inline void
 ContiguousArray<T>::Resize(std::size_t newsize) {
   if (!owned_buffer_) {
-    throw std::runtime_error("Cannot resize when using a foreign buffer; clone first");
+    throw Error("Cannot resize when using a foreign buffer; clone first");
   }
   if (newsize > capacity_) {
     std::size_t newcapacity = capacity_;
@@ -193,7 +196,7 @@ template <typename T>
 inline void
 ContiguousArray<T>::Resize(std::size_t newsize, T t) {
   if (!owned_buffer_) {
-    throw std::runtime_error("Cannot resize when using a foreign buffer; clone first");
+    throw Error("Cannot resize when using a foreign buffer; clone first");
   }
   std::size_t oldsize = Size();
   Resize(newsize);
@@ -206,7 +209,7 @@ template <typename T>
 inline void
 ContiguousArray<T>::Clear() {
   if (!owned_buffer_) {
-    throw std::runtime_error("Cannot clear when using a foreign buffer; clone first");
+    throw Error("Cannot clear when using a foreign buffer; clone first");
   }
   Resize(0);
 }
@@ -215,7 +218,7 @@ template <typename T>
 inline void
 ContiguousArray<T>::PushBack(T t) {
   if (!owned_buffer_) {
-    throw std::runtime_error("Cannot add element when using a foreign buffer; clone first");
+    throw Error("Cannot add element when using a foreign buffer; clone first");
   }
   if (size_ == capacity_) {
     Reserve(capacity_ * 2);
@@ -227,7 +230,7 @@ template <typename T>
 inline void
 ContiguousArray<T>::Extend(const std::vector<T>& other) {
   if (!owned_buffer_) {
-    throw std::runtime_error("Cannot add elements when using a foreign buffer; clone first");
+    throw Error("Cannot add elements when using a foreign buffer; clone first");
   }
   if (other.empty()) {
     return;  // appending an empty vector is a no-op
@@ -263,7 +266,7 @@ template <typename T>
 inline T&
 ContiguousArray<T>::at(std::size_t idx) {
   if (idx >= Size()) {
-    throw std::runtime_error("nid out of range");
+    throw Error("nid out of range");
   }
   return buffer_[idx];
 }
@@ -272,7 +275,7 @@ template <typename T>
 inline const T&
 ContiguousArray<T>::at(std::size_t idx) const {
   if (idx >= Size()) {
-    throw std::runtime_error("nid out of range");
+    throw Error("nid out of range");
   }
   return buffer_[idx];
 }
@@ -281,7 +284,7 @@ template <typename T>
 inline T&
 ContiguousArray<T>::at(int idx) {
   if (idx < 0 || static_cast<std::size_t>(idx) >= Size()) {
-    throw std::runtime_error("nid out of range");
+    throw Error("nid out of range");
   }
   return buffer_[static_cast<std::size_t>(idx)];
 }
@@ -290,7 +293,7 @@ template <typename T>
 inline const T&
 ContiguousArray<T>::at(int idx) const {
   if (idx < 0 || static_cast<std::size_t>(idx) >= Size()) {
-    throw std::runtime_error("nid out of range");
+    throw Error("nid out of range");
   }
   return buffer_[static_cast<std::size_t>(idx)];
 }
@@ -343,7 +346,7 @@ inline const char* InferFormatString() {
       return (std::is_unsigned<T>::value ? "=L" : "=l");
     } else {
       if (!std::is_floating_point<T>::value) {
-        throw std::runtime_error("Could not infer format string");
+        throw Error("Could not infer format string");
       }
       return "=f";
     }
@@ -352,12 +355,12 @@ inline const char* InferFormatString() {
       return (std::is_unsigned<T>::value ? "=Q" : "=q");
     } else {
       if (!std::is_floating_point<T>::value) {
-        throw std::runtime_error("Could not infer format string");
+        throw Error("Could not infer format string");
       }
       return "=d";
     }
   default:
-    throw std::runtime_error("Unrecognized type");
+    throw Error("Unrecognized type");
   }
   return nullptr;
 }
@@ -405,7 +408,7 @@ inline PyBufferFrame GetPyBufferFromScalar(T* scalar) {
 template <typename T>
 inline void InitArrayFromPyBuffer(ContiguousArray<T>* vec, PyBufferFrame frame) {
   if (sizeof(T) != frame.itemsize) {
-    throw std::runtime_error("Incorrect itemsize");
+    throw Error("Incorrect itemsize");
   }
   vec->UseForeignBuffer(frame.buf, frame.nitem);
 }
@@ -413,10 +416,10 @@ inline void InitArrayFromPyBuffer(ContiguousArray<T>* vec, PyBufferFrame frame)
 inline void InitScalarFromPyBuffer(TypeInfo* scalar, PyBufferFrame buffer) {
   using T = std::underlying_type<TypeInfo>::type;
   if (sizeof(T) != buffer.itemsize) {
-    throw std::runtime_error("Incorrect itemsize");
+    throw Error("Incorrect itemsize");
   }
   if (buffer.nitem != 1) {
-    throw std::runtime_error("nitem must be 1 for a scalar");
+    throw Error("nitem must be 1 for a scalar");
   }
   T* t = static_cast<T*>(buffer.buf);
   *scalar = static_cast<TypeInfo>(*t);
@@ -425,10 +428,10 @@ inline void InitScalarFromPyBuffer(TypeInfo* scalar, PyBufferFrame buffer) {
 inline void InitScalarFromPyBuffer(TaskType* scalar, PyBufferFrame buffer) {
   using T = std::underlying_type<TaskType>::type;
   if (sizeof(T) != buffer.itemsize) {
-    throw std::runtime_error("Incorrect itemsize");
+    throw Error("Incorrect itemsize");
   }
   if (buffer.nitem != 1) {
-    throw std::runtime_error("nitem must be 1 for a scalar");
+    throw Error("nitem must be 1 for a scalar");
   }
   T* t = static_cast<T*>(buffer.buf);
   *scalar = static_cast<TaskType>(*t);
@@ -438,10 +441,10 @@ template <typename T>
 inline void InitScalarFromPyBuffer(T* scalar, PyBufferFrame buffer) {
   static_assert(std::is_standard_layout<T>::value, "T must be in the standard layout");
   if (sizeof(T) != buffer.itemsize) {
-    throw std::runtime_error("Incorrect itemsize");
+    throw Error("Incorrect itemsize");
   }
   if (buffer.nitem != 1) {
-    throw std::runtime_error("nitem must be 1 for a scalar");
+    throw Error("nitem must be 1 for a scalar");
   }
   T* t = static_cast<T*>(buffer.buf);
   *scalar = *t;
@@ -451,7 +454,7 @@ template <typename T>
 inline void ReadScalarFromFile(T* scalar, FILE* fp) {
   static_assert(std::is_standard_layout<T>::value, "T must be in the standard layout");
   if (std::fread(scalar, sizeof(T), 1, fp) < 1) {
-    throw std::runtime_error("Could not read a scalar");
+    throw Error("Could not read a scalar");
   }
 }
 
@@ -459,7 +462,7 @@ template <typename T>
 inline void WriteScalarToFile(T* scalar, FILE* fp) {
   static_assert(std::is_standard_layout<T>::value, "T must be in the standard layout");
   if (std::fwrite(scalar, sizeof(T), 1, fp) < 1) {
-    throw std::runtime_error("Could not write a scalar");
+    throw Error("Could not write a scalar");
   }
 }
 
@@ -467,7 +470,7 @@ template <typename T>
 inline void ReadArrayFromFile(ContiguousArray<T>* vec, FILE* fp) {
   uint64_t nelem;
   if (std::fread(&nelem, sizeof(nelem), 1, fp) < 1) {
-    throw std::runtime_error("Could not read the number of elements");
+    throw Error("Could not read the number of elements");
   }
   vec->Clear();
   vec->Resize(nelem);
@@ -476,7 +479,7 @@ inline void ReadArrayFromFile(ContiguousArray<T>* vec, FILE* fp) {
   }
   const auto nelem_size_t = static_cast<std::size_t>(nelem);
   if (std::fread(vec->Data(), sizeof(T), nelem_size_t, fp) < nelem_size_t) {
-    throw std::runtime_error("Could not read an array");
+    throw Error("Could not read an array");
   }
 }
 
@@ -485,17 +488,35 @@ inline void WriteArrayToFile(ContiguousArray<T>* vec, FILE* fp) {
   static_assert(sizeof(uint64_t) >= sizeof(size_t), "size_t too large");
   const auto nelem = static_cast<uint64_t>(vec->Size());
   if (std::fwrite(&nelem, sizeof(nelem), 1, fp) < 1) {
-    throw std::runtime_error("Could not write the number of elements");
+    throw Error("Could not write the number of elements");
   }
   if (nelem == 0) {
     return;  // handle empty arrays
   }
   const auto nelem_size_t = vec->Size();
   if (std::fwrite(vec->Data(), sizeof(T), nelem_size_t, fp) < nelem_size_t) {
-    throw std::runtime_error("Could not write an array");
+    throw Error("Could not write an array");
   }
 }
 
+inline void SkipOptFieldInFile(FILE* fp) {
+  uint16_t elem_size;
+  uint64_t nelem;
+  ReadScalarFromFile(&elem_size, fp);
+  ReadScalarFromFile(&nelem, fp);
+
+  const uint64_t nbytes = elem_size * nelem;
+  TREELITE_CHECK_LE(nbytes, std::numeric_limits<long>::max());  // NOLINT
+  if (std::fseek(fp, static_cast<long>(nbytes), SEEK_CUR) != 0) {  // NOLINT
+    throw Error("Reached end of file");
+  }
+}
+
+template <typename ThresholdType, typename LeafOutputType>
+Tree<ThresholdType, LeafOutputType>::Tree(bool use_opt_field)
+  : use_opt_field_(use_opt_field)
+{}
+
 template <typename ThresholdType, typename LeafOutputType>
 inline Tree<ThresholdType, LeafOutputType>
 Tree<ThresholdType, LeafOutputType>::Clone() const {
@@ -520,8 +541,6 @@ Tree<ThresholdType, LeafOutputType>::GetFormatStringForNode() {
   }
 }
 
-constexpr std::size_t kNumFramePerTree = 7;
-
 template <typename ThresholdType, typename LeafOutputType>
 template <typename ScalarHandler, typename PrimitiveArrayHandler, typename CompositeArrayHandler>
 inline void
@@ -529,29 +548,61 @@ Tree<ThresholdType, LeafOutputType>::SerializeTemplate(
     ScalarHandler scalar_handler, PrimitiveArrayHandler primitive_array_handler,
     CompositeArrayHandler composite_array_handler) {
   scalar_handler(&num_nodes);
+  scalar_handler(&has_categorical_split_);
   composite_array_handler(&nodes_, GetFormatStringForNode());
   primitive_array_handler(&leaf_vector_);
   primitive_array_handler(&leaf_vector_begin_);
   primitive_array_handler(&leaf_vector_end_);
   primitive_array_handler(&matching_categories_);
   primitive_array_handler(&matching_categories_offset_);
+
+  /* Extension slot 2: Per-tree optional fields -- to be added later */
+  num_opt_field_per_tree_ = 0;
+  scalar_handler(&num_opt_field_per_tree_);
+
+  /* Extension slot 3: Per-node optional fields -- to be added later */
+  num_opt_field_per_node_ = 0;
+  scalar_handler(&num_opt_field_per_node_);
 }
 
 template <typename ThresholdType, typename LeafOutputType>
-template <typename ScalarHandler, typename ArrayHandler>
+template <typename ScalarHandler, typename ArrayHandler, typename SkipOptFieldHandlerFunc>
 inline void
 Tree<ThresholdType, LeafOutputType>::DeserializeTemplate(
-    ScalarHandler scalar_handler, ArrayHandler array_handler) {
+    ScalarHandler scalar_handler,
+    ArrayHandler array_handler,
+    SkipOptFieldHandlerFunc skip_opt_field_handler) {
   scalar_handler(&num_nodes);
+  scalar_handler(&has_categorical_split_);
   array_handler(&nodes_);
   if (static_cast<std::size_t>(num_nodes) != nodes_.Size()) {
-    throw std::runtime_error("Could not load the correct number of nodes");
+    throw Error("Could not load the correct number of nodes");
   }
   array_handler(&leaf_vector_);
   array_handler(&leaf_vector_begin_);
   array_handler(&leaf_vector_end_);
   array_handler(&matching_categories_);
   array_handler(&matching_categories_offset_);
+
+  if (use_opt_field_) {
+    /* Extension slot 2: Per-tree optional fields -- to be added later */
+    scalar_handler(&num_opt_field_per_tree_);
+    // Ignore extra fields; the input is likely from a later version of Treelite
+    for (int32_t i = 0; i < num_opt_field_per_tree_; ++i) {
+      skip_opt_field_handler();
+    }
+
+    /* Extension slot 3: Per-node optional fields -- to be added later */
+    scalar_handler(&num_opt_field_per_node_);
+    // Ignore extra fields; the input is likely from a later version of Treelite
+    for (int32_t i = 0; i < num_opt_field_per_node_; ++i) {
+      skip_opt_field_handler();
+    }
+  } else {
+    // Legacy (version 2.4)
+    num_opt_field_per_tree_ = 0;
+    num_opt_field_per_node_ = 0;
+  }
 }
 
 template <typename ThresholdType, typename LeafOutputType>
@@ -585,19 +636,20 @@ Tree<ThresholdType, LeafOutputType>::SerializeToFile(FILE* dest_fp) {
 }
 
 template <typename ThresholdType, typename LeafOutputType>
-inline void
-Tree<ThresholdType, LeafOutputType>::InitFromPyBuffer(std::vector<PyBufferFrame>::iterator begin,
-                                                      std::vector<PyBufferFrame>::iterator end) {
-  if (std::distance(begin, end) != kNumFramePerTree) {
-    throw std::runtime_error("Wrong number of frames specified");
-  }
-  auto scalar_handler = [&begin](auto* field) {
-    InitScalarFromPyBuffer(field, *begin++);
+inline std::vector<PyBufferFrame>::iterator
+Tree<ThresholdType, LeafOutputType>::InitFromPyBuffer(std::vector<PyBufferFrame>::iterator it) {
+  std::vector<PyBufferFrame>::iterator new_it = it;
+  auto scalar_handler = [&new_it](auto* field) {
+    InitScalarFromPyBuffer(field, *(new_it++));
   };
-  auto array_handler = [&begin](auto* field) {
-    InitArrayFromPyBuffer(field, *begin++);
+  auto array_handler = [&new_it](auto* field) {
+    InitArrayFromPyBuffer(field, *(new_it++));
   };
-  DeserializeTemplate(scalar_handler, array_handler);
+  auto skip_opt_field_handler = [&new_it]() {
+    ++new_it;
+  };
+  DeserializeTemplate(scalar_handler, array_handler, skip_opt_field_handler);
+  return new_it;
 }
 
 template <typename ThresholdType, typename LeafOutputType>
@@ -609,7 +661,10 @@ Tree<ThresholdType, LeafOutputType>::DeserializeFromFile(FILE* src_fp) {
   auto array_handler = [src_fp](auto* field) {
     ReadArrayFromFile(field, src_fp);
   };
-  DeserializeTemplate(scalar_handler, array_handler);
+  auto skip_opt_field_handler = [src_fp]() {
+    SkipOptFieldInFile(src_fp);
+  };
+  DeserializeTemplate(scalar_handler, array_handler, skip_opt_field_handler);
 }
 
 template <typename ThresholdType, typename LeafOutputType>
@@ -632,7 +687,7 @@ inline int
 Tree<ThresholdType, LeafOutputType>::AllocNode() {
   int nd = num_nodes++;
   if (nodes_.Size() != static_cast<std::size_t>(nd)) {
-    throw std::runtime_error("Invariant violated: nodes_ contains incorrect number of nodes");
+    throw Error("Invariant violated: nodes_ contains incorrect number of nodes");
   }
   for (int nid = nd; nid < num_nodes; ++nid) {
     leaf_vector_begin_.PushBack(0);
@@ -648,6 +703,7 @@ template <typename ThresholdType, typename LeafOutputType>
 inline void
 Tree<ThresholdType, LeafOutputType>::Init() {
   num_nodes = 1;
+  has_categorical_split_ = false;
   leaf_vector_.Clear();
   leaf_vector_begin_.Resize(1, {});
   leaf_vector_end_.Resize(1, {});
@@ -667,42 +723,13 @@ Tree<ThresholdType, LeafOutputType>::AddChilds(int nid) {
   nodes_.at(nid).cright_ = cright;
 }
 
-template <typename ThresholdType, typename LeafOutputType>
-inline std::vector<unsigned>
-Tree<ThresholdType, LeafOutputType>::GetCategoricalFeatures() const {
-  std::unordered_map<unsigned, bool> tmp;
-  for (int nid = 0; nid < num_nodes; ++nid) {
-    const SplitFeatureType type = SplitType(nid);
-    if (type != SplitFeatureType::kNone) {
-      const bool flag = (type == SplitFeatureType::kCategorical);
-      const uint32_t split_index = SplitIndex(nid);
-      if (tmp.count(split_index) == 0) {
-        tmp[split_index] = flag;
-      } else {
-        if (tmp[split_index] != flag) {
-          throw std::runtime_error("Feature " + std::to_string(split_index) +
-                                   " cannot be simultaneously be categorical and numerical.");
-        }
-      }
-    }
-  }
-  std::vector<unsigned> result;
-  for (const auto& kv : tmp) {
-    if (kv.second) {
-      result.push_back(kv.first);
-    }
-  }
-  std::sort(result.begin(), result.end());
-  return result;
-}
-
 template <typename ThresholdType, typename LeafOutputType>
 inline void
 Tree<ThresholdType, LeafOutputType>::SetNumericalSplit(
     int nid, unsigned split_index, ThresholdType threshold, bool default_left, Operator cmp) {
   Node& node = nodes_.at(nid);
   if (split_index >= ((1U << 31U) - 1)) {
-    throw std::runtime_error("split_index too big");
+    throw Error("split_index too big");
   }
   if (default_left) split_index |= (1U << 31U);
   node.sindex_ = split_index;
@@ -718,22 +745,22 @@ Tree<ThresholdType, LeafOutputType>::SetCategoricalSplit(
     int nid, unsigned split_index, bool default_left,
     const std::vector<uint32_t>& categories_list, bool categories_list_right_child) {
   if (split_index >= ((1U << 31U) - 1)) {
-    throw std::runtime_error("split_index too big");
+    throw Error("split_index too big");
   }
 
   const std::size_t end_oft = matching_categories_offset_.Back();
   const std::size_t new_end_oft = end_oft + categories_list.size();
   if (end_oft != matching_categories_.Size()) {
-    throw std::runtime_error("Invariant violated");
+    throw Error("Invariant violated");
   }
   if (!std::all_of(&matching_categories_offset_.at(nid + 1), matching_categories_offset_.End(),
                    [end_oft](std::size_t x) { return (x == end_oft); })) {
-    throw std::runtime_error("Invariant violated");
+    throw Error("Invariant violated");
   }
   // Hopefully we won't have to move any element as we add node_matching_categories for node nid
   matching_categories_.Extend(categories_list);
   if (new_end_oft != matching_categories_.Size()) {
-    throw std::runtime_error("Invariant violated");
+    throw Error("Invariant violated");
   }
   std::for_each(&matching_categories_offset_.at(nid + 1), matching_categories_offset_.End(),
                 [new_end_oft](std::size_t& x) { x = new_end_oft; });
@@ -746,6 +773,8 @@ Tree<ThresholdType, LeafOutputType>::SetCategoricalSplit(
   node.sindex_ = split_index;
   node.split_type_ = SplitFeatureType::kCategorical;
   node.categories_list_right_child_ = categories_list_right_child;
+
+  has_categorical_split_ = true;
 }
 
 template <typename ThresholdType, typename LeafOutputType>
@@ -824,6 +853,9 @@ Model::Dispatch(Func func) const {
 template <typename HeaderPrimitiveFieldHandlerFunc>
 inline void
 Model::SerializeTemplate(HeaderPrimitiveFieldHandlerFunc header_primitive_field_handler) {
+  major_ver_ = TREELITE_VER_MAJOR;
+  minor_ver_ = TREELITE_VER_MINOR;
+  patch_ver_ = TREELITE_VER_PATCH;
   header_primitive_field_handler(&major_ver_);
   header_primitive_field_handler(&minor_ver_);
   header_primitive_field_handler(&patch_ver_);
@@ -834,19 +866,20 @@ Model::SerializeTemplate(HeaderPrimitiveFieldHandlerFunc header_primitive_field_
 template <typename HeaderPrimitiveFieldHandlerFunc>
 inline void
 Model::DeserializeTemplate(HeaderPrimitiveFieldHandlerFunc header_primitive_field_handler,
+                           int32_t& major_ver, int32_t& minor_ver, int32_t& patch_ver,
                            TypeInfo& threshold_type, TypeInfo& leaf_output_type) {
-  int major_ver, minor_ver, patch_ver;
   header_primitive_field_handler(&major_ver);
   header_primitive_field_handler(&minor_ver);
   header_primitive_field_handler(&patch_ver);
-  if (major_ver != TREELITE_VER_MAJOR || minor_ver != TREELITE_VER_MINOR) {
+  if (major_ver != TREELITE_VER_MAJOR && !(major_ver == 2 && minor_ver == 4)) {
     std::ostringstream oss;
-    oss << "Cannot deserialize model from a different version of Treelite." << std::endl
+    oss << "Cannot deserialize model from a different major Treelite version or "
+        << "a version before 2.4.0." << std::endl
         << "Currently running Treelite version " << TREELITE_VER_MAJOR << "."
         << TREELITE_VER_MINOR << "." << TREELITE_VER_PATCH << std::endl
         << "The model checkpoint was generated from Treelite version " << major_ver << "."
         << minor_ver << "." << patch_ver;
-    throw std::runtime_error(oss.str());
+    throw Error(oss.str());
   }
   header_primitive_field_handler(&threshold_type);
   header_primitive_field_handler(&leaf_output_type);
@@ -868,6 +901,10 @@ ModelImpl<ThresholdType, LeafOutputType>::SerializeTemplate(
   header_composite_field_handler(
       &param, "T{" _TREELITE_STR(TREELITE_MAX_PRED_TRANSFORM_LENGTH) "s=f=f=f}");
 
+  /* Extension Slot 1: Per-model optional fields -- to be added later */
+  num_opt_field_per_model_ = 0;
+  header_primitive_field_handler(&num_opt_field_per_model_);
+
   /* Body */
   for (Tree<ThresholdType, LeafOutputType>& tree : trees) {
     tree_handler(tree);
@@ -875,22 +912,38 @@ ModelImpl<ThresholdType, LeafOutputType>::SerializeTemplate(
 }
 
 template <typename ThresholdType, typename LeafOutputType>
-template <typename HeaderFieldHandlerFunc, typename TreeHandlerFunc>
+template <typename HeaderFieldHandlerFunc, typename TreeHandlerFunc,
+    typename SkipOptFieldHandlerFunc>
 inline void
 ModelImpl<ThresholdType, LeafOutputType>::DeserializeTemplate(
     std::size_t num_tree,
     HeaderFieldHandlerFunc header_field_handler,
-    TreeHandlerFunc tree_handler) {
+    TreeHandlerFunc tree_handler,
+    SkipOptFieldHandlerFunc skip_opt_field_handler) {
   /* Header */
   header_field_handler(&num_feature);
   header_field_handler(&task_type);
   header_field_handler(&average_tree_output);
   header_field_handler(&task_param);
   header_field_handler(&param);
+
+  /* Extension Slot 1: Per-model optional fields -- to be added later */
+  const bool use_opt_field = (major_ver_ >= 3);
+  if (use_opt_field) {
+    header_field_handler(&num_opt_field_per_model_);
+    // Ignore extra fields; the input is likely from a later version of Treelite
+    for (int32_t i = 0; i < num_opt_field_per_model_; ++i) {
+      skip_opt_field_handler();
+    }
+  } else {
+    // Legacy (version 2.4)
+    num_opt_field_per_model_ = 0;
+  }
+
   /* Body */
   trees.clear();
   for (std::size_t i = 0; i < num_tree; ++i) {
-    trees.emplace_back();
+    trees.emplace_back(use_opt_field);
     tree_handler(trees.back());
   }
 }
@@ -898,6 +951,7 @@ ModelImpl<ThresholdType, LeafOutputType>::DeserializeTemplate(
 template <typename ThresholdType, typename LeafOutputType>
 inline void
 ModelImpl<ThresholdType, LeafOutputType>::GetPyBuffer(std::vector<PyBufferFrame>* dest) {
+  num_tree_ = static_cast<uint64_t>(this->trees.size());
   auto header_primitive_field_handler = [dest](auto* field) {
     dest->push_back(GetPyBufferFromScalar(field));
   };
@@ -907,14 +961,14 @@ ModelImpl<ThresholdType, LeafOutputType>::GetPyBuffer(std::vector<PyBufferFrame>
   auto tree_handler = [dest](Tree<ThresholdType, LeafOutputType>& tree) {
     tree.GetPyBuffer(dest);
   };
+  header_primitive_field_handler(&num_tree_);
   SerializeTemplate(header_primitive_field_handler, header_composite_field_handler, tree_handler);
 }
 
 template <typename ThresholdType, typename LeafOutputType>
 inline void
 ModelImpl<ThresholdType, LeafOutputType>::SerializeToFileImpl(FILE* dest_fp) {
-  const auto num_tree = static_cast<uint64_t>(this->trees.size());
-  WriteScalarToFile(&num_tree, dest_fp);
+  num_tree_ = static_cast<uint64_t>(this->trees.size());
   auto header_primitive_field_handler = [dest_fp](auto* field) {
     WriteScalarToFile(field, dest_fp);
   };
@@ -924,49 +978,64 @@ ModelImpl<ThresholdType, LeafOutputType>::SerializeToFileImpl(FILE* dest_fp) {
   auto tree_handler = [dest_fp](Tree<ThresholdType, LeafOutputType>& tree) {
     tree.SerializeToFile(dest_fp);
   };
+  header_primitive_field_handler(&num_tree_);
   SerializeTemplate(header_primitive_field_handler, header_composite_field_handler, tree_handler);
 }
 
 template <typename ThresholdType, typename LeafOutputType>
-inline void
+inline std::vector<PyBufferFrame>::iterator
 ModelImpl<ThresholdType, LeafOutputType>::InitFromPyBuffer(
-    std::vector<PyBufferFrame>::iterator begin, std::vector<PyBufferFrame>::iterator end) {
-  const std::size_t num_frame = std::distance(begin, end);
-  constexpr std::size_t kNumFrameInHeader = 5;
-  if (num_frame < kNumFrameInHeader || (num_frame - kNumFrameInHeader) % kNumFramePerTree != 0) {
-    throw std::runtime_error("Wrong number of frames");
+    std::vector<PyBufferFrame>::iterator it, std::size_t num_frame) {
+  std::vector<PyBufferFrame>::iterator new_it = it;
+  auto header_field_handler = [&new_it](auto* field) {
+    InitScalarFromPyBuffer(field, *(new_it++));
+  };
+
+  auto skip_opt_field_handler = [&new_it]() {
+    ++new_it;
+  };
+
+  auto tree_handler = [&new_it](Tree<ThresholdType, LeafOutputType>& tree) {
+    new_it = tree.InitFromPyBuffer(new_it);
+  };
+
+  if (major_ver_ == 2) {
+    // From version 2.4 (legacy)
+    // num_tree has to be inferred from the number of frames received
+    constexpr std::size_t kNumFrameInHeader = 5;
+    constexpr std::size_t kNumFramePerTree = 8;
+    num_tree_ = (num_frame - kNumFrameInHeader) / kNumFramePerTree;
+  } else {
+    // From version 3.x
+    // num_tree is now explicitly stored
+    header_field_handler(&num_tree_);
   }
-  const std::size_t num_tree = (num_frame - kNumFrameInHeader) / kNumFramePerTree;
 
-  auto header_field_handler = [&begin](auto* field) {
-    InitScalarFromPyBuffer(field, *begin++);
-  };
+  DeserializeTemplate(num_tree_, header_field_handler, tree_handler, skip_opt_field_handler);
+  TREELITE_CHECK_EQ(num_tree_, this->trees.size());
 
-  auto tree_handler = [&begin](Tree<ThresholdType, LeafOutputType>& tree) {
-    // Read the frames in the range [begin, begin + kNumFramePerTree) into the tree
-    tree.InitFromPyBuffer(begin, begin + kNumFramePerTree);
-    begin += kNumFramePerTree;
-      // Advance the iterator so that the next tree reads the next kNumFramePerTree frames
-  };
-
-  DeserializeTemplate(num_tree, header_field_handler, tree_handler);
+  return new_it;
 }
 
 template <typename ThresholdType, typename LeafOutputType>
 inline void
 ModelImpl<ThresholdType, LeafOutputType>::DeserializeFromFileImpl(FILE* src_fp) {
-  uint64_t num_tree;
-  ReadScalarFromFile(&num_tree, src_fp);
+  ReadScalarFromFile(&num_tree_, src_fp);
 
   auto header_field_handler = [src_fp](auto* field) {
     ReadScalarFromFile(field, src_fp);
   };
 
+  auto skip_opt_field_handler = [src_fp]() {
+    SkipOptFieldInFile(src_fp);
+  };
+
   auto tree_handler = [src_fp](Tree<ThresholdType, LeafOutputType>& tree) {
     tree.DeserializeFromFile(src_fp);
   };
 
-  DeserializeTemplate(num_tree, header_field_handler, tree_handler);
+  DeserializeTemplate(num_tree_, header_field_handler, tree_handler, skip_opt_field_handler);
+  TREELITE_CHECK_EQ(num_tree_, this->trees.size());
 }
 
 inline void InitParamAndCheck(ModelParam* param,
