@@ -58,6 +58,7 @@ class ASTNativeCompilerImpl {
     task_param_ = model.task_param;
     pred_transform_ = model.param.pred_transform;
     sigmoid_alpha_ = model.param.sigmoid_alpha;
+    ratio_c_ = model.param.ratio_c;
     global_bias_ = model.param.global_bias;
     files_.clear();
 
@@ -147,6 +148,7 @@ class ASTNativeCompilerImpl {
   TaskParam task_param_;
   std::string pred_transform_;
   float sigmoid_alpha_;
+  float ratio_c_;
   float global_bias_;
   std::string pred_tranform_func_;
   std::string array_is_categorical_;
@@ -224,6 +226,7 @@ class ASTNativeCompilerImpl {
           "num_feature"_a = num_feature_,
           "pred_transform"_a = pred_transform_,
           "sigmoid_alpha"_a = sigmoid_alpha_,
+          "ratio_c"_a = ratio_c_,
           "global_bias"_a = global_bias_,
           "threshold_type_str"_a = TypeInfoToString(TypeToInfo<ThresholdType>()),
           "leaf_output_type_str"_a = TypeInfoToString(TypeToInfo<LeafOutputType>()));
@@ -639,6 +642,11 @@ class ASTNativeCompilerImpl {
           "split_index"_a = node->split_index,
           "right_categories_flag"_a = right_categories_flag);
       }
+
+      oss << fmt::format(
+          "((data[{split_index}].fvalue >= 0) && "
+          "(fabsf(data[{split_index}].fvalue) <= (float)(1U << FLT_MANT_DIG)) && (",
+          "split_index"_a = node->split_index);
       oss << "(tmp >= 0 && tmp < 64 && (( (uint64_t)"
           << bitmap[0] << "U >> tmp) & 1) )";
       for (size_t i = 1; i < bitmap.size(); ++i) {
@@ -647,7 +655,7 @@ class ASTNativeCompilerImpl {
             << " && (( (uint64_t)" << bitmap[i]
             << "U >> (tmp - " << (i * 64) << ") ) & 1) )";
       }
-      oss << ")";
+      oss << ")))";
       result = oss.str();
     }
     return result;
